# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import click

from bentoml.deployment.serverless import ServerlessDeployment
from bentoml.deployment.sagemaker import SagemakerDeployment
from bentoml.cli.click_utils import _echo, CLI_COLOR_ERROR, CLI_COLOR_SUCCESS
from bentoml.utils.usage_stats import track_cli

SERVERLESS_PLATFORMS = ["aws-lambda", "aws-lambda-py2", "gcp-function"]

# pylint: disable=unused-variable


def add_deployment_commands(cli):

    # Example usage: bentoml deploy /ARCHIVE_PATH --platform=aws-lambda
    @cli.command(
        help="Deploy BentoML archive as REST endpoint to cloud services",
        short_help="Deploy Bento archive",
    )
    @click.argument("archive-path", type=click.STRING)
    @click.option(
        "--platform",
        type=click.Choice(
            [
                "aws-lambda",
                "aws-lambda-py2",
                "gcp-function",
                "aws-sagemaker",
                "azure-ml",
                "algorithmia",
            ]
        ),
        required=True,
        help="Target platform that Bento archive is going to deployed to",
    )
    @click.option(
        "--region",
        type=click.STRING,
        help="Target region inside the cloud provider that will be deployed to",
    )
    @click.option("--stage", type=click.STRING)
    @click.option(
        "--api-name", type=click.STRING, help="The name of API will be deployed"
    )
    @click.option(
        "--instance-type",
        type=click.STRING,
        help="SageMaker deployment ONLY. The instance type to use for deployment",
    )
    @click.option(
        "--instance-count",
        type=click.INT,
        help="Sagemaker deployment ONLY. Number of instances to use for deployment",
    )
    def deploy(
        archive_path, platform, region, stage, api_name, instance_type, instance_count
    ):
        track_cli('deploy', platform)
        if platform in SERVERLESS_PLATFORMS:
            deployment = ServerlessDeployment(archive_path, platform, region, stage)
        elif platform == "aws-sagemaker":
            deployment = SagemakerDeployment(
                archive_path, api_name, region, instance_count, instance_type
            )
        else:
            _echo(
                "Deploying with --platform=%s is not supported in current version of "
                "BentoML" % platform,
                CLI_COLOR_ERROR,
            )
            return

        output_path = deployment.deploy()

        _echo(
            "Successfully deployed to {platform}!".format(platform=platform),
            CLI_COLOR_SUCCESS,
        )
        _echo("Deployment archive is saved at: %s" % output_path)

    # Example usage: bentoml delete-deployment ARCHIVE_PATH --platform=aws-lambda
    @cli.command(
        help="Delete active BentoML deployment from cloud services",
        short_help="Delete active BentoML deployment",
    )
    @click.argument("archive-path", type=click.STRING)
    @click.option(
        "--platform",
        type=click.Choice(
            [
                "aws-lambda",
                "aws-lambda-py2",
                "gcp-function",
                "aws-sagemaker",
                "azure-ml",
                "algorithmia",
            ]
        ),
        required=True,
        help="The platform bento archive is deployed to",
    )
    @click.option(
        "--region",
        type=click.STRING,
        required=True,
        help="The region deployment belongs to",
    )
    @click.option(
        "--api-name",
        type=click.STRING,
        help="Name of the API function that is deployed",
    )
    @click.option("--stage", type=click.STRING)
    def delete_deployment(archive_path, platform, region, stage, api_name):
        track_cli('delete-deploy', platform)
        if platform in SERVERLESS_PLATFORMS:
            deployment = ServerlessDeployment(archive_path, platform, region, stage)
        elif platform == "aws-sagemaker":
            deployment = SagemakerDeployment(archive_path, api_name, region)
        else:
            _echo(
                "Remove deployment with --platform=%s is not supported in current "
                "version of BentoML" % platform,
                CLI_COLOR_ERROR,
            )
            return

        if deployment.delete():
            _echo(
                "Successfully delete {platform} deployment".format(platform=platform),
                CLI_COLOR_SUCCESS,
            )
        else:
            _echo(
                "Delete {platform} deployment unsuccessful".format(platform=platform),
                CLI_COLOR_ERROR,
            )

    # Example usage: bentoml check-deployment-status ARCHIVE_PATH --platform=aws-lambda
    @cli.command(
        help="Check deployment status of BentoML archive",
        short_help="check deployment status",
    )
    @click.argument("archive-path", type=click.STRING)
    @click.option(
        "--platform",
        type=click.Choice(
            [
                "aws-lambda",
                "aws-lambda-py2",
                "gcp-function",
                "aws-sagemaker",
                "azure-ml",
                "algorithmia",
            ]
        ),
        required=True,
        help="Target platform that Bento archive will be deployed to as a REST api \
                service",
    )
    @click.option(
        "--region",
        type=click.STRING,
        required=True,
        help="Deployment's region name inside cloud provider.",
    )
    @click.option("--stage", type=click.STRING)
    @click.option(
        "--api-name",
        type=click.STRING,
        help="The name of API that is deployed as a service.",
    )
    def check_deployment_status(archive_path, platform, region, stage, api_name):
        track_cli('check-deployment-status', platform)
        if platform in SERVERLESS_PLATFORMS:
            deployment = ServerlessDeployment(archive_path, platform, region, stage)
        elif platform == "aws-sagemaker":
            deployment = SagemakerDeployment(archive_path, api_name, region)
        else:
            _echo(
                "check deployment status with --platform=%s is not supported in the "
                "current version of BentoML" % platform,
                CLI_COLOR_ERROR,
            )
            return

        deployment.check_status()

    return cli



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from bentoml.utils import cloudpickle
from bentoml.artifact import ArtifactSpec, ArtifactInstance

try:
    import tensorflow as tf
    from tensorflow.python import keras
except ImportError:
    tf = None
    keras = None


class KerasModelArtifact(ArtifactSpec):
    """
    Abstraction for saving/loading Keras model
    """

    def __init__(self, name, custom_objects=None, model_extension=".h5"):
        super(KerasModelArtifact, self).__init__(name)
        self._model_extension = model_extension

        self.custom_objects = custom_objects
        self.graph = None
        self.sess = None

    def _custom_objects_path(self, base_path):
        return os.path.join(base_path, self.name + '_custom_objects.pkl')

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + self._model_extension)

    def bind_keras_backend_session(self):
        if tf is None:
            raise ImportError(
                "Tensorflow package is required to use KerasModelArtifact."
            )

        self.sess = tf.compat.v1.keras.backend.get_session()
        self.graph = self.sess.graph

    def creat_session(self):
        if tf is None:
            raise ImportError(
                "Tensorflow package is required to use KerasModelArtifact."
            )

        self.graph = tf.compat.v1.get_default_graph()
        self.sess = tf.Session(graph=self.graph)
        tf.keras.backend.set_session(self.sess)

    def pack(self, data):  # pylint:disable=arguments-differ
        if tf is None:
            raise ImportError(
                "Tensorflow package is required to use KerasModelArtifact."
            )

        if isinstance(data, keras.engine.training.Model):
            model = data
            custom_objects = self.custom_objects
        elif (
            isinstance(data, dict)
            and 'model' in data
            and isinstance(data['model'], keras.engine.training.Model)
        ):
            model = data['model']
            custom_objects = (
                data['custom_objects']
                if 'custom_objects' in data
                else self.custom_objects
            )
        else:
            raise ValueError(
                "KerasModelArtifact#pack expects type: keras.engine.training.Model"
            )

        self.bind_keras_backend_session()
        model._make_predict_function()
        return _TfKerasModelArtifactInstance(self, model, custom_objects)

    def load(self, path):
        if tf is None:
            raise ImportError(
                "Tensorflow package is required to use KerasModelArtifact."
            )

        self.creat_session()

        if self.custom_objects is None and os.path.isfile(
            self._custom_objects_path(path)
        ):
            self.custom_objects = cloudpickle.load(
                open(self._custom_objects_path(path), 'rb')
            )

        with self.graph.as_default():
            with self.sess.as_default():
                model = keras.models.load_model(
                    self._model_file_path(path), custom_objects=self.custom_objects
                )
        return self.pack(model)


class _TfKerasModelArtifactInstance(ArtifactInstance):
    def __init__(self, spec, model, custom_objects):
        super(_TfKerasModelArtifactInstance, self).__init__(spec)

        if tf is None:
            raise ImportError(
                "Tensorflow package is required to use KerasModelArtifact."
            )

        if not isinstance(model, keras.engine.training.Model):
            raise ValueError("Expected `model` argument to be a "
                             "`keras.engine.training.Model` instance")

        self.graph = spec.graph
        self.sess = spec.sess
        self._model = model
        self._custom_objects = custom_objects
        self._model_wrapper = _TfKerasModelWrapper(self._model, self.graph, self.sess)

    def save(self, dst):
        # save custom_objects for model
        cloudpickle.dump(
            self._custom_objects, open(self.spec._custom_objects_path(dst), "wb")
        )

        # save keras model
        self._model.save(self.spec._model_file_path(dst))

    def get(self):
        return self._model_wrapper


class _TfKerasModelWrapper:
    def __init__(self, keras_model, graph, sess):
        self.keras_model = keras_model
        self.graph = graph
        self.sess = sess

    def predict(self, *args, **kwargs):
        with self.graph.as_default():
            with self.sess.as_default():
                return self.keras_model.predict(*args, **kwargs)

    def predict_classes(self, *args, **kwargs):
        with self.graph.as_default():
            with self.sess.as_default():
                return self.keras_model.predict_classes(*args, **kwargs)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
import base64
import logging
import re
from six.moves.urllib.parse import urlparse

import boto3
import docker

from bentoml.deployment.base_deployment import Deployment
from bentoml.deployment.utils import (
    generate_bentoml_deployment_snapshot_path,
    process_docker_api_line,
)
from bentoml.utils.whichcraft import which
from bentoml.exceptions import BentoMLException
from bentoml.deployment.sagemaker.templates import (
    DEFAULT_NGINX_CONFIG,
    DEFAULT_WSGI_PY,
    DEFAULT_SERVE_SCRIPT,
)

logger = logging.getLogger(__name__)

DEFAULT_REGION = "us-west-2"
DEFAULT_INSTANCE_TYPE = "ml.m4.xlarge"
DEFAULT_INSTANCE_COUNT = 1


def strip_scheme(url):
    """ Stripe url's schema
    e.g.   http://some.url/path -> some.url/path
    :param url: String
    :return: String
    """
    parsed = urlparse(url)
    scheme = "%s://" % parsed.scheme
    return parsed.geturl().replace(scheme, "", 1)


def generate_aws_compatible_string(item):
    pattern = re.compile("[^a-zA-Z0-9-]|_")
    return re.sub(pattern, "-", item)


def get_arn_role_from_current_user():
    sts_client = boto3.client("sts")
    identity = sts_client.get_caller_identity()
    sts_arn = identity["Arn"]
    sts_arn_list = sts_arn.split(":")
    type_role = sts_arn_list[-1].split("/")
    iam_client = boto3.client("iam")
    if type_role[0] == "user":
        role_list = iam_client.list_roles()
        arn = None
        for role in role_list["Roles"]:
            policy_document = role["AssumeRolePolicyDocument"]
            statement = policy_document["Statement"][0]
            if (
                statement["Effect"] == "Allow"
                and statement["Principal"]["Service"] == "sagemaker.amazonaws.com"
            ):
                arn = role["Arn"]
        if arn is None:
            raise ValueError(
                "Can't find proper Arn role for Sagemaker, please create one and try "
                "again"
            )
        return arn
    elif type_role[0] == "role":
        role_response = iam_client.get_role(RoleName=type_role[1])
        return role_response["Role"]["Arn"]


def create_push_image_to_ecr(bento_service, snapshot_path):
    """Create BentoService sagemaker image and push to AWS ECR

    Example: https://github.com/awslabs/amazon-sagemaker-examples/blob/\
        master/advanced_functionality/scikit_bring_your_own/container/build_and_push.sh
    1. get aws account info and login ecr
    2. create ecr repository, if not exist
    3. build tag and push docker image

    :param bento_service: Bento Service
    :param snapshot_path: Path
    :return: ecr_tag: String
    """
    ecr_client = boto3.client("ecr")
    token = ecr_client.get_authorization_token()
    logger.info("Getting docker login info from AWS")
    username, password = (
        base64.b64decode(token["authorizationData"][0]["authorizationToken"])
        .decode("utf-8")
        .split(":")
    )
    registry_url = token["authorizationData"][0]["proxyEndpoint"]
    auth_config_payload = {"username": username, "password": password}

    docker_api = docker.APIClient()

    image_name = bento_service.name.lower() + "-sagemaker"
    ecr_tag = strip_scheme(
        "{registry_url}/{image_name}:{version}".format(
            registry_url=registry_url,
            image_name=image_name,
            version=bento_service.version,
        )
    )

    logger.info("Building docker image: %s", image_name)
    for line in docker_api.build(
        path=snapshot_path, dockerfile="Dockerfile-sagemaker", tag=image_name
    ):
        process_docker_api_line(line)

    try:
        ecr_client.describe_repositories(repositoryNames=[image_name])["repositories"]
    except ecr_client.exceptions.RepositoryNotFoundException:
        ecr_client.create_repository(repositoryName=image_name)

    if docker_api.tag(image_name, ecr_tag) is False:
        raise RuntimeError("Tag appeared to fail: " + ecr_tag)
    logger.info("Pushing image to AWS ECR at %s", ecr_tag)
    for line in docker_api.push(ecr_tag, stream=True, auth_config=auth_config_payload):
        process_docker_api_line(line)
    logger.info("Finished pushing image: %s", ecr_tag)
    return ecr_tag


class SagemakerDeployment(Deployment):
    def __init__(
        self,
        archive_path,
        api_name,
        region=None,
        instance_count=None,
        instance_type=None,
    ):
        if which("docker") is None:
            raise ValueError(
                "docker is not installed, please install docker and then try again"
            )
        super(SagemakerDeployment, self).__init__(archive_path)
        self.region = DEFAULT_REGION if region is None else region
        self.instance_count = (
            DEFAULT_INSTANCE_COUNT if instance_count is None else instance_count
        )
        self.instant_type = (
            DEFAULT_INSTANCE_TYPE if instance_type is None else instance_type
        )
        apis = self.bento_service.get_service_apis()
        if api_name:
            self.api = next(item for item in apis if item.name == api_name)
        elif len(apis) == 1:
            self.api = apis[0]
        else:
            raise BentoMLException(
                "Please specify api-name, when more than one API is present in the "
                "archive"
            )
        self.sagemaker_client = boto3.client("sagemaker", region_name=self.region)
        self.model_name = generate_aws_compatible_string(
            "bentoml-" + self.bento_service.name + "-" + self.bento_service.version
        )
        self.endpoint_config_name = generate_aws_compatible_string(
            self.bento_service.name
            + "-"
            + self.bento_service.version
            + "-configuration"
        )

    def deploy(self):
        """Deploy BentoML service to AWS Sagemaker.
        Your AWS credential must have the correct permissions for sagemaker and ECR

        1. Create docker image and push to ECR
        2. Create sagemaker model base on the ECR image
        3. Create sagemaker endpoint configuration base on sagemaker model
        4. Create sagemaker endpoint base on sagemaker endpoint configuration

        :param archive_path: Path
        :param additional_info: Dict
        :return: String, location to the output snapshot's path
        """
        snapshot_path = generate_bentoml_deployment_snapshot_path(
            self.bento_service.name, self.bento_service.version, "aws-sagemaker"
        )
        shutil.copytree(self.archive_path, snapshot_path)
        with open(os.path.join(snapshot_path, "nginx.conf"), "w") as f:
            f.write(DEFAULT_NGINX_CONFIG)
        with open(os.path.join(snapshot_path, "wsgi.py"), "w") as f:
            f.write(DEFAULT_WSGI_PY)
        with open(os.path.join(snapshot_path, "serve"), "w") as f:
            f.write(DEFAULT_SERVE_SCRIPT)

        # permission 755 is required for entry script 'serve'
        permission = "755"
        octal_permission = int(permission, 8)
        os.chmod(os.path.join(snapshot_path, "serve"), octal_permission)

        execution_role_arn = get_arn_role_from_current_user()
        ecr_image_path = create_push_image_to_ecr(self.bento_service, snapshot_path)
        sagemaker_model_info = {
            "ModelName": self.model_name,
            "PrimaryContainer": {
                "ContainerHostname": self.model_name,
                "Image": ecr_image_path,
                "Environment": {"API_NAME": self.api.name},
            },
            "ExecutionRoleArn": execution_role_arn,
        }
        logger.info("Creating sagemaker model %s", self.model_name)
        create_model_response = self.sagemaker_client.create_model(
            **sagemaker_model_info
        )
        logger.info("AWS create model response: %s", create_model_response)

        production_variants = [
            {
                "VariantName": self.bento_service.name,
                "ModelName": self.model_name,
                "InitialInstanceCount": self.instance_count,
                "InstanceType": self.instant_type,
            }
        ]
        logger.info(
            "Creating sagemaker endpoint %s configuration", self.endpoint_config_name
        )
        create_endpoint_config_response = self.sagemaker_client.create_endpoint_config(
            EndpointConfigName=self.endpoint_config_name,
            ProductionVariants=production_variants,
        )
        logger.info(
            "AWS create endpoint config response: %s", create_endpoint_config_response
        )

        logger.info("Creating sagemaker endpoint %s", self.bento_service.name)
        create_endpoint_response = self.sagemaker_client.create_endpoint(
            EndpointName=self.bento_service.name,
            EndpointConfigName=self.endpoint_config_name,
        )
        logger.info("AWS create endpoint response: %s", create_endpoint_response)

        # TODO: maybe wait for this endpoint from creating to running and then return
        return snapshot_path

    def check_status(self):
        endpoint_status_response = self.sagemaker_client.describe_endpoint(
            EndpointName=self.bento_service.name
        )
        logger.info("AWS describe endpoint response: %s", endpoint_status_response)
        endpoint_in_service = endpoint_status_response["EndpointStatus"] == "InService"

        status_message = "{service} is {status}".format(
            service=self.bento_service.name,
            status=endpoint_status_response["EndpointStatus"],
        )
        if endpoint_in_service:
            status_message += (
                "\nEndpoint ARN: " + endpoint_status_response["EndpointArn"]
            )

        return endpoint_in_service, status_message

    def delete(self):
        """Delete Sagemaker endpoint for the bentoml service.
        It will also delete the model or the endpoint configuration.

        return: Boolean, True if the deletion is successful
        """
        if not self.check_status()[0]:
            raise BentoMLException(
                "No active AWS Sagemaker deployment for service %s"
                % self.bento_service.name
            )

        delete_endpoint_response = self.sagemaker_client.delete_endpoint(
            EndpointName=self.bento_service.name
        )
        logger.info("AWS delete endpoint response: %s", delete_endpoint_response)
        if delete_endpoint_response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            # We will also try to delete both model and endpoint configuration for user.
            # Since they are not critical, even they failed, we will still count delete
            # deployment a success action
            delete_model_response = self.sagemaker_client.delete_model(
                ModelName=self.model_name
            )
            logger.info("AWS delete model response: %s", delete_model_response)
            if delete_model_response["ResponseMetadata"]["HTTPStatusCode"] != 200:
                logger.error(
                    "Encounter error when deleting model: %s", delete_model_response
                )

            delete_endpoint_config_response = self.sagemaker_client.delete_endpoint_config(  # noqa: E501
                EndpointConfigName=self.endpoint_config_name
            )
            logger.info(
                "AWS delete endpoint config response: %s",
                delete_endpoint_config_response,
            )
            if (
                delete_endpoint_config_response["ResponseMetadata"]["HTTPStatusCode"]
                != 200
            ):
                logger.error(
                    "Encounter error when deleting endpoint configuration: %s",
                    delete_endpoint_config_response,
                )
            return True
        else:
            return False



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import argparse
import base64
from io import BytesIO

from werkzeug.utils import secure_filename
from flask import Response
import numpy as np

from bentoml.exceptions import BentoMLException
from bentoml.handlers.base_handlers import BentoHandler, get_output_str
from bentoml.handlers.image_handler import ACCEPTED_CONTENT_TYPES, check_file_format


class FastaiImageHandler(BentoHandler):
    """Transform incoming image data to fastai.vision.Image

    Handle incoming image data, process them into fastai.vision.Image instance and
    pass down to user defined API functions


    Args:
        input_name ([str]]): A list of acceptable input name for HTTP request.
            Default value is image
        accept_file_extensions ([str]):  A list of acceptable image extensions.
            Default value is [.jpg, .jpeg, .png]
        accept_multiple_files (boolean):  Accept multiple files in single request or
            not. Default value is False
        convert_mode (str): The pilmode to be used for reading image file into
            numpy array. Default value is RGB.  Find more information at
            https://imageio.readthedocs.io/en/stable/format_png-pil.html#png-pil
        div (bool): If True, pixel values are divided by 255 to become floats
            between 0. and 1.
        cls (Class): Parameter from fastai.vision ``open_image``, default is
            ``fastai.vision.Image``
        after_open (func): Parameter from fastai.vision ``open_image``, default
            is None

    Raises:
        ImportError: imageio package is required to use FastaiImageHandler
        ImportError: fastai package is required to use FastaiImageHandler
    """

    def __init__(
        self,
        input_name=None,
        accept_file_extensions=None,
        convert_mode=None,
        div=True,
        cls=None,
        after_open=None,
    ):
        self.input_name = input_name or "image"
        self.convert_mode = convert_mode or "RGB"
        self.div = (div or True,)
        self.cls = cls
        self.accept_file_extensions = accept_file_extensions or [".jpg", ".png", "jpeg"]
        self.after_open = after_open

    @property
    def request_schema(self):
        return {
            "image/*": {"schema": {"type": "string", "format": "binary"}},
            "multipart/form-data": {
                "schema": {"type": "object"},
                "properties": {self.input_name: {"type": "string", "format": "binary"}},
            },
        }

    @property
    def pip_dependencies(self):
        return ['imageio', 'fastai']

    def handle_request(self, request, func):
        try:
            from fastai.vision import Image, pil2tensor
        except ImportError:
            raise ImportError("fastai package is required to use FastaiImageHandler")

        try:
            from imageio import imread
        except ImportError:
            raise ImportError("imageio package is required to use FastaiImageHandler")

        if request.method != "POST":
            return Response(response="Only accept POST request", status=400)

        input_file = request.files.get(self.input_name)

        if input_file:
            file_name = secure_filename(input_file.filename)
            check_file_format(file_name, self.accept_file_extensions)
            input_stream = BytesIO(input_file.read())
        elif request.data:
            input_stream = request.data
        else:
            raise ValueError(
                "BentoML#FastaiImageHandler unexpected HTTP request: %s" % request
            )

        input_data = imread(input_stream, pilmode=self.convert_mode)

        if self.after_open:
            input_data = self.after_open(input_data)

        input_data = pil2tensor(input_data, np.float32)

        if self.div:
            input_data = input_data.div_(255)

        if self.cls:
            input_data = self.cls(input_data)
        else:
            input_data = Image(input_data)

        result = func(input_data)
        result = get_output_str(result, request.headers.get("output", "json"))
        return Response(response=result, status=200, mimetype="application/json")

    def handle_cli(self, args, func):
        parser = argparse.ArgumentParser()
        parser.add_argument("--input", required=True)
        parser.add_argument(
            "-o", "--output", default="str", choices=["str", "json", "yaml"]
        )
        parsed_args = parser.parse_args(args)
        file_path = parsed_args.input

        check_file_format(file_path, self.accept_file_extensions)
        if not os.path.isabs(file_path):
            file_path = os.path.abspath(file_path)

        try:
            from fastai.vision import open_image, Image
        except ImportError:
            raise ImportError("fastai package is required to use")

        image_array = open_image(
            fn=file_path,
            convert_mode=self.convert_mode,
            div=self.div,
            after_open=self.after_open,
            cls=self.cls or Image,
        )

        result = func(image_array)
        result = get_output_str(result, output_format=parsed_args.output)
        print(result)

    def handle_aws_lambda_event(self, event, func):
        try:
            from imageio import imread
        except ImportError:
            raise ImportError("imageio package is required to use ImageHandler")

        if event["headers"].get("Content-Type", None) in ACCEPTED_CONTENT_TYPES:
            # decodebytes introduced at python3.1
            try:
                image_data = imread(
                    base64.decodebytes(event["body"]), pilmode=self.pilmode
                )
            except AttributeError:
                image_data = imread(
                    base64.decodestring(event["body"]),  # pylint: disable=W1505
                    pilmode=self.convert_mode,
                )
        else:
            raise BentoMLException(
                "BentoML currently doesn't support Content-Type: {content_type} for "
                "AWS Lambda".format(content_type=event["headers"]["Content-Type"])
            )
        try:
            from fastai.vision import pil2tensor, Image
        except ImportError:
            raise ImportError("fastai package is required")

        if self.after_open:
            image_data = self.after_open(image_data)

        image_data = pil2tensor(image_data, np.float32)
        if self.div:
            image_data = image_data.div_(255)
        if self.cls:
            image_data = self.cls(image_data)
        else:
            image_data = Image(image_data)

        result = func(image_data)
        result = get_output_str(result, event["headers"].get("output", "json"))
        return {"statusCode": 200, "body": result}

    def handle_clipper_bytes(self, inputs, func):
        try:
            import cv2
        except ImportError:
            raise ImportError(
                "opencv-python package is required to use FastaiImageHandler"
            )

        try:
            from fastai.vision import pil2tensor, Image
        except ImportError:
            raise ImportError("fastai package is required to use")

        def transform_and_predict(input_bytes):
            image_data = cv2.imdecode(input_bytes, cv2.IMREAD_COLOR)
            if self.after_open:
                image_data = self.after_open(image_data)
            image_data = pil2tensor(image_data, np.float32)

            if self.div:
                image_data = image_data.div_(255)

            if self.cls:
                image_data = self.cls(image_data)
            else:
                image_data = Image(image_data)

            return func(image_data)

        return list(map(transform_and_predict, inputs))

    def handle_clipper_strings(self, inputs, func):
        raise RuntimeError(
            "Fastai Image handler does not support 'strings' input_type \
                for Clipper deployment at the moment"
        )

    def handle_clipper_ints(self, inputs, func):
        raise RuntimeError(
            "Fastai ImageHandler doesn't support ints input types \
                for clipper deployment at the moment"
        )

    def handle_clipper_doubles(self, inputs, func):
        raise RuntimeError(
            "Fastai ImageHandler doesn't support doubles input types \
                for clipper deployment at the moment"
        )

    def handle_clipper_floats(self, inputs, func):
        raise RuntimeError(
            "Fastai ImageHandler doesn't support floats input types \
                for clipper deployment at the moment"
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import click
import shutil
import logging

from configparser import ConfigParser

from bentoml.config import (
    config as bentoml_config,
    LOCAL_CONFIG_FILE,
    DEFAULT_CONFIG_FILE,
)
from bentoml.cli.click_utils import _echo, CLI_COLOR_ERROR

# pylint: disable=unused-variable

LOG = logging.getLogger(__name__)

EXAMPLE_CONFIG_USAGE = '''
Example usage for `bentoml config`:
  bentoml config set usage_tracking=false
  bentoml config set apiserver.default_port=9000
  bentoml config unset apiserver.default_port
'''


def create_local_config_file_if_not_found():
    if not os.path.isfile(LOCAL_CONFIG_FILE):
        LOG.info("Creating new default BentoML config file at: %s", LOCAL_CONFIG_FILE)
        shutil.copyfile(DEFAULT_CONFIG_FILE, LOCAL_CONFIG_FILE)


def add_configuration_commands(cli):
    @cli.command(
        help="Setting local default BentoML configurations",
        short_help="Config BentoML library",
        context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
    )
    @click.argument(
        "action", type=click.Choice(["view", "view-effective", "set", "unset", "reset"])
    )
    @click.argument("updates", nargs=-1)
    def config(action, updates):
        create_local_config_file_if_not_found()

        if action == "view-effective":
            bentoml_config.write(sys.stdout)
            return

        if action == "reset":
            if os.path.isfile(LOCAL_CONFIG_FILE):
                LOG.info("Removing existing BentoML config file: %s", LOCAL_CONFIG_FILE)
                os.remove(LOCAL_CONFIG_FILE)
            create_local_config_file_if_not_found()
            return

        local_config = ConfigParser()
        with open(LOCAL_CONFIG_FILE, 'rb') as config_file:
            local_config.read_string(config_file.read().decode('utf-8'))

        if action == "view":
            local_config.write(sys.stdout)
        elif action == "set":
            try:
                for update in updates:
                    item, value = update.split('=')
                    if '.' in item:
                        sec, opt = item.split('.')
                    else:
                        sec = 'core'  # default section
                        opt = item

                    if not local_config.has_section(sec):
                        local_config.add_section(sec)
                    local_config.set(sec.strip(), opt.strip(), value.strip())

                local_config.write(open(LOCAL_CONFIG_FILE, 'w'))

            except ValueError:
                _echo('Wrong config format: %s' % str(updates), CLI_COLOR_ERROR)
                _echo(EXAMPLE_CONFIG_USAGE)
                return

        elif action == "unset":
            try:
                for update in updates:
                    if '.' in update:
                        sec, opt = update.split('.')
                    else:
                        sec = 'core'  # default section
                        opt = update

                    if not local_config.has_section(sec):
                        local_config.add_section(sec)
                    local_config.remove_option(sec.strip(), opt.strip())

                local_config.write(open(LOCAL_CONFIG_FILE, 'w'))

            except ValueError:
                _echo('Wrong config format: %s' % str(updates), CLI_COLOR_ERROR)
                _echo(EXAMPLE_CONFIG_USAGE)
                return
        else:
            LOG.error("Unknown command: bentoml config %s", action)
            return
        return

    return cli



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import os
import uuid
import logging
from datetime import datetime

from bentoml.utils import Path
from bentoml.utils.s3 import is_s3_url, upload_to_s3
from bentoml.utils.tempdir import TempDirectory
from bentoml.archive.py_module_utils import copy_used_py_modules
from bentoml.archive.templates import (
    BENTO_MODEL_SETUP_PY_TEMPLATE,
    MANIFEST_IN_TEMPLATE,
    BENTO_SERVICE_DOCKERFILE_CPU_TEMPLATE,
    BENTO_SERVICE_DOCKERFILE_SAGEMAKER_TEMPLATE,
    INIT_PY_TEMPLATE,
)
from bentoml.archive.config import BentoArchiveConfig

DEFAULT_BENTO_ARCHIVE_DESCRIPTION = """\
# BentoML(bentoml.ai) generated model archive
"""

LOG = logging.getLogger(__name__)


def _validate_version_str(version_str):
    """
    Validate that version str format:
    * Consist of only ALPHA / DIGIT / "-" / "." / "_"
    * Length between 1-128
    """
    regex = r"[A-Za-z0-9_.-]{1,128}\Z"
    if re.match(regex, version_str) is None:
        raise ValueError(
            'Invalid BentoArchive version: "{}", it can only consist'
            ' ALPHA / DIGIT / "-" / "." / "_", and must be less than'
            "128 characthers".format(version_str)
        )


def _generate_new_version_str():
    """
    Generate a version string in the format of YYYY-MM-DD-Hash
    """
    time_obj = datetime.now()
    date_string = time_obj.strftime("%Y_%m_%d")
    random_hash = uuid.uuid4().hex[:8]

    return date_string + "_" + random_hash


def save(bento_service, dst, version=None):
    """Save given BentoService along with all its artifacts, source code and
    dependencies to target path

    Args:
        bento_service (bentoml.service.BentoService): a Bento Service instance
        dst (str): Destination of where the bento service will be saved. It could
            be a local file path or a s3 path
        version (:obj:`str`, optional): version text to use for saved archive

    Returns:
        string: The complete path of saved Bento service.
    """

    if version is None:
        version = _generate_new_version_str()
    _validate_version_str(version)

    if (
        bento_service._version_major is not None
        and bento_service._version_minor is not None
    ):
        # BentoML uses semantic versioning for BentoService distribution
        # when user specified the MAJOR and MINOR version number along with
        # the BentoService class definition with '@ver' decorator.
        # The parameter version(or auto generated version) here will be used as
        # PATCH field in the final version:
        version = ".".join(
            [
                str(bento_service._version_major),
                str(bento_service._version_minor),
                version,
            ]
        )

    # Full path containing saved BentoArchive, it the dst path with service name
    # and service version as prefix. e.g.:
    # - s3://my-bucket/base_path => s3://my-bucket/base_path/service_name/version/
    # - /tmp/my_bento_archive/ => /tmp/my_bento_archive/service_name/version/
    full_saved_path = os.path.join(dst, bento_service.name, version)

    if is_s3_url(dst):
        with TempDirectory() as tempdir:
            _save(bento_service, tempdir, version)
            upload_to_s3(full_saved_path, tempdir)
    else:
        _save(bento_service, dst, version)

    LOG.info(
        "BentoService %s:%s saved to %s", bento_service.name, version, full_saved_path
    )
    return full_saved_path


def _save(bento_service, dst, version=None):
    Path(os.path.join(dst), bento_service.name).mkdir(parents=True, exist_ok=True)
    # Update path to subfolder in the form of 'base/service_name/version/'
    path = os.path.join(dst, bento_service.name, version)

    if os.path.exists(path):
        raise ValueError(
            "Version {version} in Path: {dst} already "
            "exist.".format(version=version, dst=dst)
        )

    os.mkdir(path)
    module_base_path = os.path.join(path, bento_service.name)
    os.mkdir(module_base_path)

    # write README.md with user model's docstring
    if bento_service.__class__.__doc__:
        model_description = bento_service.__class__.__doc__.strip()
    else:
        model_description = DEFAULT_BENTO_ARCHIVE_DESCRIPTION
    with open(os.path.join(path, "README.md"), "w") as f:
        f.write(model_description)

    # save all model artifacts to 'base_path/name/artifacts/' directory
    bento_service.artifacts.save(module_base_path)

    # write conda environment, requirement.txt
    bento_service.env.save(path)

    # TODO: add bentoml.find_packages helper for more fine grained control over this
    # process, e.g. packages=find_packages(base, [], exclude=[], used_module_only=True)
    # copy over all custom model code
    module_name, module_file = copy_used_py_modules(
        bento_service.__class__.__module__, os.path.join(path, bento_service.name)
    )

    # create __init__.py
    with open(os.path.join(path, bento_service.name, "__init__.py"), "w") as f:
        f.write(
            INIT_PY_TEMPLATE.format(
                service_name=bento_service.name,
                module_name=module_name,
                pypi_package_version=version,
            )
        )

    # write setup.py, make exported model pip installable
    setup_py_content = BENTO_MODEL_SETUP_PY_TEMPLATE.format(
        name=bento_service.name,
        pypi_package_version=version,
        long_description=model_description,
    )
    with open(os.path.join(path, "setup.py"), "w") as f:
        f.write(setup_py_content)

    with open(os.path.join(path, "MANIFEST.in"), "w") as f:
        f.write(MANIFEST_IN_TEMPLATE.format(service_name=bento_service.name))

    # write Dockerfile
    with open(os.path.join(path, "Dockerfile"), "w") as f:
        f.write(BENTO_SERVICE_DOCKERFILE_CPU_TEMPLATE)
    with open(os.path.join(path, "Dockerfile-sagemaker"), "w") as f:
        f.write(BENTO_SERVICE_DOCKERFILE_SAGEMAKER_TEMPLATE)

    # write bentoml.yml
    config = BentoArchiveConfig()
    config["metadata"].update(
        {
            "service_name": bento_service.name,
            "service_version": version,
            "module_name": module_name,
            "module_file": module_file,
        }
    )
    config["env"] = bento_service.env.to_dict()

    config.write_to_path(path)
    # Also write bentoml.yml to module base path to make it accessible
    # as package data after pip installed as a python package
    config.write_to_path(module_base_path)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import click
import logging

from bentoml.archive import load
from bentoml.server import BentoAPIServer, get_docs
from bentoml.server.gunicorn_server import GunicornBentoServer
from bentoml.cli.click_utils import DefaultCommandGroup, conditional_argument
from bentoml.cli.deployment import add_deployment_commands
from bentoml.cli.config import add_configuration_commands
from bentoml.utils.log import configure_logging
from bentoml.utils.usage_stats import track_cli


def create_bento_service_cli(archive_path=None):
    # pylint: disable=unused-variable

    @click.group(cls=DefaultCommandGroup)
    @click.option(
        '-q',
        '--quiet',
        is_flag=True,
        help="Hide process logs and only print command results",
    )
    @click.option(
        '--debug', is_flag=True, help="Print debugging info for BentoML developer"
    )
    @click.version_option()
    @click.pass_context
    def bentoml_cli(ctx, debug, quiet):
        """
        BentoML CLI tool
        """
        ctx.debug = debug
        ctx.quiet = quiet

        if debug:
            configure_logging(logging.DEBUG)
        elif quiet:
            configure_logging(logging.WARNING)
        else:
            configure_logging()  # use default setting in local bentoml.cfg

    # Example Usage: bentoml API_NAME /SAVED_ARCHIVE_PATH --input=INPUT
    @bentoml_cli.command(
        default_command=True,
        default_command_usage="API_NAME BENTO_ARCHIVE_PATH --input=INPUT",
        default_command_display_name="<API_NAME>",
        short_help="Run API function",
        help="Run a API defined in saved BentoArchive with cli args as input",
        context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
    )
    @click.argument("api-name", type=click.STRING)
    @conditional_argument(archive_path is None, "archive-path", type=click.STRING)
    @click.pass_context
    def run(ctx, api_name, archive_path=archive_path):
        track_cli('run')
        bento_service = load(archive_path)

        try:
            api = next(
                (
                    api
                    for api in bento_service.get_service_apis()
                    if api.name == api_name
                )
            )
        except StopIteration:
            raise ValueError(
                "Can't find API '{}' in Service '{}'".format(
                    api_name, bento_service.name
                )
            )

        api.handle_cli(ctx.args)

    # Example Usage: bentoml info /SAVED_ARCHIVE_PATH
    @bentoml_cli.command(
        help="List all APIs defined in the BentoService loaded from archive.",
        short_help="List APIs",
    )
    @conditional_argument(archive_path is None, "archive-path", type=click.STRING)
    def info(archive_path=archive_path):
        """
        List all APIs defined in the BentoService loaded from archive
        """
        track_cli('info')
        bento_service = load(archive_path)

        service_apis = bento_service.get_service_apis()
        output = json.dumps(
            dict(
                name=bento_service.name,
                version=bento_service.version,
                apis=[api.name for api in service_apis],
            ),
            indent=2,
        )
        print(output)

    # Example usage: bentoml docs /SAVED_ARCHIVE_PATH
    @bentoml_cli.command(
        help="Display API documents in Open API format",
        short_help="Display OpenAPI docs",
    )
    @conditional_argument(archive_path is None, "archive-path", type=click.STRING)
    def docs(archive_path=archive_path):
        track_cli('docs')
        bento_service = load(archive_path)

        print(json.dumps(get_docs(bento_service), indent=2))

    # Example Usage: bentoml serve ./SAVED_ARCHIVE_PATH --port=PORT
    @bentoml_cli.command(
        help="Start REST API server hosting BentoService loaded from archive",
        short_help="Start local rest server",
    )
    @conditional_argument(archive_path is None, "archive-path", type=click.STRING)
    @click.option(
        "--port",
        type=click.INT,
        default=BentoAPIServer._DEFAULT_PORT,
        help="The port to listen on for the REST api server, default is 5000.",
    )
    def serve(port, archive_path=archive_path):
        track_cli('serve')
        bento_service = load(archive_path)

        server = BentoAPIServer(bento_service, port=port)
        server.start()

    # Example Usage:
    # bentoml serve-gunicorn ./SAVED_ARCHIVE_PATH --port=PORT --workers=WORKERS
    @bentoml_cli.command(
        help="Start REST API gunicorn server hosting BentoService loaded from archive",
        short_help="Start local gunicorn server",
    )
    @conditional_argument(archive_path is None, "archive-path", type=click.STRING)
    @click.option("-p", "--port", type=click.INT, default=None)
    @click.option(
        "-w",
        "--workers",
        type=click.INT,
        default=None,
        help="Number of workers will start for the gunicorn server",
    )
    @click.option("--timeout", type=click.INT, default=None)
    def serve_gunicorn(port, workers, timeout, archive_path=archive_path):
        track_cli('serve_gunicorn')

        gunicorn_app = GunicornBentoServer(archive_path, port, workers, timeout)
        gunicorn_app.run()

    # pylint: enable=unused-variable
    return bentoml_cli


def create_bentoml_cli():
    _cli = create_bento_service_cli()

    # Commands created here aren't mean to be used from generated service archive. They
    # are used as part of BentoML cli commands only.

    add_deployment_commands(_cli)
    add_configuration_commands(_cli)

    # pylint: enable=unused-variable
    return _cli


cli = create_bentoml_cli()

if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import argparse
import base64
from io import BytesIO

from werkzeug.utils import secure_filename
from flask import Response

from bentoml.exceptions import BentoMLException
from bentoml.handlers.base_handlers import BentoHandler, get_output_str

ACCEPTED_CONTENT_TYPES = ["images/png", "images/jpeg", "images/jpg"]


def check_file_format(file_name, accept_format_list):
    """
    Raise error if file's extension is not in the accept_format_list
    """
    if accept_format_list:
        _, extension = os.path.splitext(file_name)
        if extension not in accept_format_list:
            raise ValueError(
                "Input file not in supported format list: {}".format(accept_format_list)
            )


class ImageHandler(BentoHandler):
    """Transform incoming image data from http request, cli or lambda event into numpy
    array.

    Handle incoming image data from different sources, transform them into numpy array
    and pass down to user defined API functions

    Args:
        input_name (string[]]): A list of acceptable input name for HTTP request.
            Default value is image
        accept_file_extensions (string[]):  A list of acceptable image extensions.
            Default value is [.jpg, .jpeg, .png]
        accept_multiple_files (boolean):  Accept multiple files in single request or
            not. Default value is False
        pilmode (string): The pilmode to be used for reading image file into numpy
            array. Default value is RGB.  Find more information at
            https://imageio.readthedocs.io/en/stable/format_png-pil.html#png-pil

    Raises:
        ImportError: imageio package is required to use ImageHandler
    """

    def __init__(
        self,
        input_name="image",
        accept_file_extensions=None,
        accept_multiple_files=False,
        pilmode="RGB",
    ):
        self.input_name = input_name
        self.pilmode = pilmode
        self.accept_file_extensions = accept_file_extensions or [
            ".jpg",
            ".png",
            ".jpeg",
        ]
        self.accept_multiple_files = accept_multiple_files

    @property
    def request_schema(self):
        return {
            "image/*": {"schema": {"type": "string", "format": "binary"}},
            "multipart/form-data": {
                "schema": {"type": "object"},
                "properties": {self.input_name: {"type": "string", "format": "binary"}},
            },
        }

    @property
    def pip_dependencies(self):
        return ['imageio']

    def handle_request(self, request, func):
        """Handle http request that has image file/s. It will convert image into a
        ndarray for the function to consume.

        Args:
            request: incoming request object.
            func: function that will take ndarray as its arg.
            options: configuration for handling request object.
        Return:
            response object
        """
        try:
            from imageio import imread
        except ImportError:
            raise ImportError("imageio package is required to use ImageHandler")

        if request.method != "POST":
            return Response(response="Only accept POST request", status=400)

        if not self.accept_multiple_files:
            input_file = request.files.get(self.input_name)

            if input_file:
                file_name = secure_filename(input_file.filename)
                check_file_format(file_name, self.accept_file_extensions)
                input_stream = BytesIO(input_file.read())
            elif request.data:
                input_stream = request.data
            else:
                raise ValueError(
                    "BentoML#ImageHandler unexpected HTTP request: %s" % request
                )

            input_data = imread(input_stream, pilmode=self.pilmode)
        else:
            return Response(response="Only support single file input", status=400)

        result = func(input_data)
        result = get_output_str(result, request.headers.get("output", "json"))
        return Response(response=result, status=200, mimetype="application/json")

    def handle_cli(self, args, func):
        parser = argparse.ArgumentParser()
        parser.add_argument("--input", required=True)
        parser.add_argument(
            "-o", "--output", default="str", choices=["str", "json", "yaml"]
        )
        parsed_args = parser.parse_args(args)
        file_path = parsed_args.input

        check_file_format(file_path, self.accept_file_extensions)
        if not os.path.isabs(file_path):
            file_path = os.path.abspath(file_path)

        try:
            from imageio import imread
        except ImportError:
            raise ImportError("imageio package is required to use ImageHandler")

        image_array = imread(file_path, pilmode=self.pilmode)

        result = func(image_array)
        result = get_output_str(result, output_format=parsed_args.output)
        print(result)

    def handle_aws_lambda_event(self, event, func):
        try:
            from imageio import imread
        except ImportError:
            raise ImportError("imageio package is required to use ImageHandler")

        if event["headers"].get("Content-Type", None) in ACCEPTED_CONTENT_TYPES:
            # decodebytes introduced at python3.1
            try:
                image = imread(base64.decodebytes(event["body"]), pilmode=self.pilmode)
            except AttributeError:
                image = imread(
                    base64.decodestring(event["body"]),  # pylint: disable=W1505
                    pilmode=self.pilmode,
                )
        else:
            raise BentoMLException(
                "BentoML currently doesn't support Content-Type: {content_type} for "
                "AWS Lambda".format(content_type=event["headers"]["Content-Type"])
            )

        result = func(image)
        result = get_output_str(result, event["headers"].get("output", "json"))
        return {"statusCode": 200, "body": result}

    def handle_clipper_bytes(self, inputs, func):
        try:
            import cv2
        except ImportError:
            raise ImportError("opencv-python package is required to use ImageHandler")

        def transform_and_predict(input_bytes):
            data = cv2.imdecode(input_bytes, cv2.IMREAD_COLOR)
            return func(data)

        return list(map(transform_and_predict, inputs))

    def handle_clipper_strings(self, inputs, func):
        raise RuntimeError(
            "ImageHandler does not support 'strings' input_type \
                for Clipper deployment at the moment"
        )

    def handle_clipper_ints(self, inputs, func):
        raise RuntimeError(
            "ImageHandler doesn't support ints input types \
                for clipper deployment at the moment"
        )

    def handle_clipper_doubles(self, inputs, func):
        raise RuntimeError(
            "ImageHandler doesn't support doubles input types \
                for clipper deployment at the moment"
        )

    def handle_clipper_floats(self, inputs, func):
        raise RuntimeError(
            "ImageHandler doesn't support floats input types \
                for clipper deployment at the moment"
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from bentoml.archive import load


class Deployment(object):
    """Deployment is spec for describing what actions deployment should have
    to interact with BentoML cli and BentoML service archive.
    """

    def __init__(self, archive_path):
        self.bento_service = load(archive_path)
        self.archive_path = archive_path

    def deploy(self):
        """Deploy bentoml service.

        :return: Boolean, True if success
        """
        raise NotImplementedError

    def check_status(self):
        """Check deployment status

        :params
        :return: Boolean, True if success Status Message String
        """
        raise NotImplementedError

    def delete(self):
        """Delete deployment, if deployment is active
        :return: Boolean, True if success
        """
        raise NotImplementedError



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from bentoml.handlers.base_handlers import BentoHandler


class PytorchTensorHandler(BentoHandler):
    """
    Tensor handlers for Pytorch models
    """

    def handle_request(self, request, func):
        raise NotImplementedError

    def handle_cli(self, args, func):
        raise NotImplementedError

    def handle_aws_lambda_event(self, event, func):
        raise NotImplementedError



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os

from ruamel.yaml import YAML

from bentoml.utils import Path

DEFAULT_AWS_REGION = "us-west-2"
DEFAULT_AWS_DEPLOY_STAGE = "dev"

logger = logging.getLogger(__name__)

AWS_HANDLER_PY_TEMPLATE_HEADER = """\
try:
    import unzip_requirements
except ImportError:
    pass

from {class_name} import {class_name}

bento_service = {class_name}.load()
apis = bento_service.get_service_apis()

"""

AWS_FUNCTION_TEMPLATE = """\
def {api_name}(event, context):
    api = next(item for item in apis if item.name == '{api_name}')

    result = api.handle_aws_lambda_event(event)
    return result

"""


def generate_serverless_configuration_for_aws(
    service_name, apis, output_path, region, stage
):
    config_path = os.path.join(output_path, "serverless.yml")
    yaml = YAML()
    with open(config_path, "r") as f:
        content = f.read()
    serverless_config = yaml.load(content)

    serverless_config["service"] = service_name
    serverless_config["provider"]["region"] = region
    logger.info("Using user AWS region: %s", region)

    serverless_config["provider"]["stage"] = stage
    logger.info("Using AWS stage: %s", stage)

    serverless_config["functions"] = {}
    for api in apis:
        function_config = {
            "handler": "handler.{name}".format(name=api.name),
            "events": [
                {"http": {"path": "/{name}".format(name=api.name), "method": "post"}}
            ],
        }
        serverless_config["functions"][api.name] = function_config

    custom_config = {
        "apigwBinary": ["image/jpg", "image/jpeg", "image/png"],
        "pythonRequirements": {
            "useDownloadCache": True,
            "useStaticCache": True,
            "dockerizePip": True,
            "slim": True,
            "strip": True,
            "zip": True,
        },
    }

    serverless_config["custom"] = custom_config

    yaml.dump(serverless_config, Path(config_path))
    return


def generate_handler_py(bento_service, apis, output_path):
    handler_py_content = AWS_HANDLER_PY_TEMPLATE_HEADER.format(
        class_name=bento_service.name
    )

    for api in apis:
        api_content = AWS_FUNCTION_TEMPLATE.format(api_name=api.name)
        handler_py_content = handler_py_content + api_content

    with open(os.path.join(output_path, "handler.py"), "w") as f:
        f.write(handler_py_content)
    return


def create_aws_lambda_bundle(bento_service, output_path, region, stage):
    apis = bento_service.get_service_apis()
    generate_handler_py(bento_service, apis, output_path)
    generate_serverless_configuration_for_aws(
        bento_service.name, apis, output_path, region, stage
    )
    return



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import logging

from bentoml.utils import Path
from bentoml.exceptions import BentoMLConfigException
from bentoml.config.configparser import BentoMLConfigParser

logger = logging.getLogger(__name__)


def expand_env_var(env_var):
    """Expands potentially nested env var by repeatedly applying `expandvars` and
    `expanduser` until interpolation stops having any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


def parameterized_config(template):
    """Generates a configuration from the provided template + variables defined in
    current scope

    Args:
        :param template: a config content templated with {{variables}}
    Returns:
        string: config content after templated with locals() and globals()
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)


BENTOML_HOME = expand_env_var(os.environ.get("BENTOML_HOME", "~/bentoml"))
try:
    Path(BENTOML_HOME).mkdir(exist_ok=True)
except OSError as err:
    raise BentoMLConfigException(
        "Error creating bentoml home dir '{}': {}".format(BENTOML_HOME, err.strerror)
    )

# Default bentoml config comes with the library bentoml/config/default_bentoml.cfg
DEFAULT_CONFIG_FILE = os.path.join(os.path.dirname(__file__), "default_bentoml.cfg")
with open(DEFAULT_CONFIG_FILE, "rb") as f:
    DEFAULT_CONFIG = f.read().decode("utf-8")

config = BentoMLConfigParser(default_config=parameterized_config(DEFAULT_CONFIG))

if "BENTML_CONFIG" in os.environ:
    # User local config file for customizing bentoml
    LOCAL_CONFIG_FILE = expand_env_var(os.environ.get("BENTML_CONFIG"))
    logger.info("Using BentoML config file $BENTML_CONFIG: %s", LOCAL_CONFIG_FILE)
    with open(LOCAL_CONFIG_FILE, "rb") as f:
        config.read_string(parameterized_config(f.read().decode("utf-8")))
else:
    LOCAL_CONFIG_FILE = os.path.join(BENTOML_HOME, "bentoml.cfg")
    if os.path.isfile(LOCAL_CONFIG_FILE):
        logger.info("Using local default BentoML config file: %s", LOCAL_CONFIG_FILE)
        with open(LOCAL_CONFIG_FILE, "rb") as f:
            config.read_string(parameterized_config(f.read().decode("utf-8")))
    else:
        logger.info("No local BentoML config file found, using default configurations")



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re

from six.moves.urllib.parse import urlparse, uses_netloc, uses_params, uses_relative

try:
    from pathlib import Path

    Path().expanduser()
except (ImportError, AttributeError):
    from pathlib2 import Path

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

_VALID_URLS = set(uses_relative + uses_netloc + uses_params)
_VALID_URLS.discard("")


def is_url(url):
    try:
        return urlparse(url).scheme in _VALID_URLS
    except Exception:  # pylint:disable=broad-except
        return False


def isidentifier(s):
    """
    Return true if string is in a valid python identifier format:

    https://docs.python.org/2/reference/lexical_analysis.html#identifiers
    """
    try:
        return s.isidentifier()
    except AttributeError:
        # str#isidentifier is only available in python 3
        return re.match(r"[A-Za-z_][A-Za-z_0-9]*\Z", s) is not None



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import sys
import platform
import json
import logging

import uuid
import requests

import bentoml


logger = logging.getLogger(__name__)

AMPLITUDE_URL = "https://api.amplitude.com/httpapi"
API_KEY = '7f65f2446427226eb86f6adfacbbf47a'

bentoml_version = bentoml.__version__
platform_info = platform.platform()
py_version = "{major}.{minor}.{micro}".format(
    major=sys.version_info.major,
    minor=sys.version_info.minor,
    micro=sys.version_info.micro,
)


def get_bento_service_info(bento_service):
    if not isinstance(bento_service, bentoml.BentoService):
        return {}

    artifact_types = []
    handler_types = []

    for artifact in bento_service.artifacts.items():
        artifact_instance = artifact[1]
        artifact_types.append(artifact_instance.__class__.__name__)

    for api in bento_service.get_service_apis():
        handler_types.append(api.handler.__class__.__name__)

    return {
        "handler_types": handler_types,
        "artifact_types": artifact_types,
        "env": bento_service.env.to_dict(),
    }


def track(event_type, info):
    info['py_version'] = py_version
    info["bento_version"] = bentoml_version
    info["platform_info"] = platform_info

    return send_amplitude_event(event_type, info)


def track_save(bento_service):
    if bentoml.config['core'].getboolean("usage_tracking"):
        info = get_bento_service_info(bento_service)
        return track("save", info)


def track_loading(bento_service):
    if bentoml.config['core'].getboolean("usage_tracking"):
        info = get_bento_service_info(bento_service)
        return track("load", info)


def track_cli(command, deploy_platform=None):
    if bentoml.config['core'].getboolean("usage_tracking"):
        info = {}
        if deploy_platform is not None:
            info['platform'] = deploy_platform
        return track('cli-' + command, info)


def send_amplitude_event(event, info):
    """Send event to amplitude
    https://developers.amplitude.com/?java#keys-for-the-event-argument
    """

    event_info = [
        {"event_type": event, "user_id": str(uuid.uuid4()), "event_properties": info}
    ]
    event_data = {"api_key": API_KEY, "event": json.dumps(event_info)}

    try:
        requests.post(AMPLITUDE_URL, data=event_data, timeout=1)
        return
    except Exception as error:  # pylint:disable=broad-except
        # silently fail
        logger.info(str(error))
        return



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import tempfile

from bentoml.utils.usage_stats import track_loading
from bentoml.utils.s3 import is_s3_url, download_from_s3
from bentoml.exceptions import BentoMLException
from bentoml.archive.config import BentoArchiveConfig


def load_bentoml_config(path):
    try:
        return BentoArchiveConfig.load(os.path.join(path, "bentoml.yml"))
    except FileNotFoundError:
        raise ValueError(
            "BentoML can't locate config file 'bentoml.yml'"
            " in archive path: {}".format(path)
        )


def load_bento_service_class(archive_path):
    """
    Load a BentoService class from saved archive in given path

    :param archive_path: A BentoArchive path generated from BentoService.save call
        or the path to pip installed BentoArchive directory
    :return: BentoService class
    """
    if is_s3_url(archive_path):
        tempdir = tempfile.mkdtemp()
        download_from_s3(archive_path, tempdir)
        archive_path = tempdir

    config = load_bentoml_config(archive_path)
    metadata = config["metadata"]

    # Load target module containing BentoService class from given path
    module_file_path = os.path.join(
        archive_path, metadata["service_name"], metadata["module_file"]
    )
    if not os.path.isfile(module_file_path):
        # Try loading without service_name prefix, for loading from a installed PyPi
        module_file_path = os.path.join(archive_path, metadata["module_file"])

    if not os.path.isfile(module_file_path):
        raise BentoMLException(
            "Can not locate module_file {} in archive {}".format(
                metadata["module_file"], archive_path
            )
        )

    # Prepend archive_path to sys.path for loading extra python dependencies
    sys.path.insert(0, archive_path)

    module_name = metadata["module_name"]
    if module_name in sys.modules:
        # module already loaded, TODO: add warning
        module = sys.modules[module_name]
    elif sys.version_info >= (3, 5):
        import importlib.util

        spec = importlib.util.spec_from_file_location(module_name, module_file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    elif sys.version_info >= (3, 3):
        from importlib.machinery import SourceFileLoader

        # pylint:disable=deprecated-method
        module = SourceFileLoader(module_name, module_file_path).load_module(
            module_name
        )
        # pylint:enable=deprecated-method
    else:
        import imp

        module = imp.load_source(module_name, module_file_path)

    # Remove archive_path from sys.path to avoid import naming conflicts
    sys.path.remove(archive_path)

    model_service_class = module.__getattribute__(metadata["service_name"])
    # Set _bento_archive_path, which tells BentoService where to load its artifacts
    model_service_class._bento_archive_path = archive_path
    # Set cls._version, service instance can access it via svc.version
    model_service_class._bento_service_version = metadata["service_version"]

    return model_service_class


def load(archive_path):
    """Load bento service from local file path or s3 path

    Args:
        archive_path (str): The path that contains archived bento service.
            It could be local file path or aws s3 path

    Returns:
        bentoml.service.BentoService: The loaded bento service.
    """
    svc_cls = load_bento_service_class(archive_path)
    svc = svc_cls.from_archive(archive_path)
    track_loading(svc)
    return svc



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import logging

from datetime import datetime

from bentoml.config import BENTOML_HOME

logger = logging.getLogger(__name__)


def generate_bentoml_deployment_snapshot_path(service_name, service_version, platform):
    return os.path.join(
        BENTOML_HOME,
        "deployment-snapshots",
        platform,
        service_name,
        service_version,
        datetime.now().isoformat(),
    )


def process_docker_api_line(payload):
    """ Process the output from API stream, throw an Exception if there is an error """
    # Sometimes Docker sends to "{}\n" blocks together...
    for segment in payload.decode("utf-8").split("\n"):
        line = segment.strip()
        if line:
            try:
                line_payload = json.loads(line)
            except ValueError as e:
                print("Could not decipher payload from Docker API: " + e)
            if line_payload:
                if "errorDetail" in line_payload:
                    error = line_payload["errorDetail"]
                    logger.error(error['message'])
                    raise RuntimeError("Error on build - code %s" % error["code"])
                elif "stream" in line_payload:
                    logger.info(line_payload['stream'])



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import logging

from ruamel.yaml import YAML

from bentoml.utils import Path

DEFAULT_GCP_REGION = "us-west2"
DEFAULT_GCP_DEPLOY_STAGE = "dev"

logger = logging.getLogger(__name__)

GOOGLE_MAIN_PY_TEMPLATE_HEADER = """\
from {class_name} import {class_name}

bento_service = {class_name}.load()
apis = bento_service.get_service_apis()

"""

GOOGLE_FUNCTION_TEMPLATE = """\
def {api_name}(request):
    api = next(item for item in apis if item.name == '{api_name}')

    result = api.handle_request(request)
    return result

"""


def generate_serverless_configuration_for_google(
    bento_service, apis, output_path, region, stage
):
    config_path = os.path.join(output_path, "serverless.yml")
    yaml = YAML()
    with open(config_path, "r") as f:
        content = f.read()
    serverless_config = yaml.load(content)

    serverless_config["service"] = bento_service.name
    serverless_config["provider"]["project"] = bento_service.name

    serverless_config["provider"]["region"] = region
    logger.info("Using user defined Google region: %s", region)

    serverless_config["provider"]["stage"] = stage
    logger.info("Using user defined Google stage: %s", stage)

    serverless_config["functions"] = {}
    for api in apis:
        function_config = {"handler": api.name, "events": [{"http": "path"}]}
        serverless_config["functions"][api.name] = function_config

    yaml.dump(serverless_config, Path(config_path))
    return


def generate_main_py(bento_service, apis, output_path):
    handler_py_content = GOOGLE_MAIN_PY_TEMPLATE_HEADER.format(
        class_name=bento_service.name
    )

    for api in apis:
        api_content = GOOGLE_FUNCTION_TEMPLATE.format(api_name=api.name)
        handler_py_content = handler_py_content + api_content

    with open(os.path.join(output_path, "main.py"), "w") as f:
        f.write(handler_py_content)
    return


def create_gcp_function_bundle(bento_service, output_path, region, stage):
    apis = bento_service.get_service_apis()
    generate_main_py(bento_service, apis, output_path)
    generate_serverless_configuration_for_google(
        bento_service, apis, output_path, region, stage
    )
    return



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
import logging
import subprocess
from subprocess import PIPE

from ruamel.yaml import YAML
from packaging import version

from bentoml.utils import Path
from bentoml.utils.tempdir import TempDirectory
from bentoml.utils.whichcraft import which
from bentoml.exceptions import BentoMLException
from bentoml.deployment.base_deployment import Deployment
from bentoml.deployment.serverless.aws_lambda_template import (
    create_aws_lambda_bundle,
    DEFAULT_AWS_DEPLOY_STAGE,
    DEFAULT_AWS_REGION,
)
from bentoml.deployment.serverless.gcp_function_template import (
    create_gcp_function_bundle,
    DEFAULT_GCP_REGION,
    DEFAULT_GCP_DEPLOY_STAGE,
)
from bentoml.deployment.utils import generate_bentoml_deployment_snapshot_path

logger = logging.getLogger(__name__)

SERVERLESS_PROVIDER = {
    "aws-lambda": "aws-python3",
    "aws-lambda-py2": "aws-python",
    "gcp-function": "google-python",
}


def check_serverless_compatiable_version():
    if which("serverless") is None:
        raise ValueError(
            "Serverless framework is not installed, please visit "
            + "www.serverless.com for install instructions."
        )

    version_result = subprocess.check_output(["serverless", "-v"])
    parsed_version = version.parse(version_result.decode("utf-8").strip())

    if parsed_version >= version.parse("1.40.0"):
        return
    else:
        raise ValueError(
            "Incompatiable serverless version, please install version 1.40.0 or greater"
        )


def parse_serverless_response(serverless_response):
    """Parse serverless response string, raise error if it is a serverless error,
    otherwise, return information.
    """
    str_list = serverless_response.strip().split("\n")
    error = [s for s in str_list if "Serverless Error" in s]
    if error:
        error_pos = str_list.index(error[0])
        error_message = str_list[error_pos + 1]
        raise BentoMLException(error_message)
    return str_list


class ServerlessDeployment(Deployment):
    """Managing deployment operations for serverless
    """

    def __init__(self, archive_path, platform, region, stage):
        check_serverless_compatiable_version()
        super(ServerlessDeployment, self).__init__(archive_path)

        self.platform = platform
        self.provider = SERVERLESS_PROVIDER[platform]
        if platform == "google-python":
            self.region = DEFAULT_GCP_REGION if region is None else region
            self.stage = DEFAULT_GCP_DEPLOY_STAGE if stage is None else stage
        elif platform == "aws-lambda" or platform == "aws-lambda-py":
            self.region = DEFAULT_AWS_REGION if region is None else region
            self.stage = DEFAULT_AWS_DEPLOY_STAGE if stage is None else stage
        else:
            raise ValueError(
                "This version of BentoML doesn't support platform %s" % platform
            )

    def _generate_bundle(self):
        output_path = generate_bentoml_deployment_snapshot_path(
            self.bento_service.name, self.bento_service.version, self.platform
        )
        Path(output_path).mkdir(parents=True, exist_ok=False)

        # Calling serverless command to generate templated project
        with subprocess.Popen(
            [
                "serverless",
                "create",
                "--template",
                self.provider,
                "--name",
                self.bento_service.name,
            ],
            cwd=output_path,
            stdout=PIPE,
            stderr=PIPE,
        ) as proc:
            response = parse_serverless_response(proc.stdout.read().decode("utf-8"))
            logger.debug("Serverless response: %s", "\n".join(response))

        if self.platform == "google-python":
            create_gcp_function_bundle(
                self.bento_service, output_path, self.region, self.stage
            )
        elif self.platform == "aws-lambda" or self.platform == "aws-lambda-py2":
            # Installing two additional plugins to make it works for AWS lambda
            # serverless-python-requirements will packaging required python modules,
            # and automatically compress and create layer
            with subprocess.Popen(
                [
                    "serverless",
                    "plugin",
                    "install",
                    "-n",
                    "serverless-python-requirements",
                ],
                cwd=output_path,
                stdout=PIPE,
                stderr=PIPE,
            ) as proc:
                response = parse_serverless_response(proc.stdout.read().decode("utf-8"))
                logger.debug("Serverless response: %s", "\n".join(response))

            with subprocess.Popen(
                ["serverless", "plugin", "install", "-n", "serverless-apigw-binary"],
                cwd=output_path,
                stdout=PIPE,
                stderr=PIPE,
            ) as proc:
                response = parse_serverless_response(proc.stdout.read().decode("utf-8"))
                logger.debug("Serverless response: %s", "\n".join(response))

            create_aws_lambda_bundle(
                self.bento_service, output_path, self.region, self.stage
            )
        else:
            raise BentoMLException(
                "%s is not supported in current version of BentoML" % self.provider
            )

        shutil.copy(os.path.join(self.archive_path, "requirements.txt"), output_path)

        model_serivce_archive_path = os.path.join(output_path, self.bento_service.name)
        shutil.copytree(self.archive_path, model_serivce_archive_path)

        return os.path.realpath(output_path)

    def deploy(self):
        output_path = self._generate_bundle()
        with subprocess.Popen(
            ["serverless", "deploy"], cwd=output_path, stdout=PIPE, stderr=PIPE
        ) as proc:
            response = parse_serverless_response(proc.stdout.read().decode("utf-8"))
            logger.debug("Serverless response: %s", "\n".join(response))
            service_info_index = response.index("Service Information")
            service_info = response[service_info_index:]
            logger.info("BentoML: %s", "\n".join(service_info))
            print("\n".join(service_info))
            return output_path

    def check_status(self):
        """Check deployment status for the bentoml service.
        return True, if it is active else return false
        """

        apis = self.bento_service.get_service_apis()
        config = {
            "service": self.bento_service.name,
            "provider": {"region": self.region, "stage": self.stage},
            "functions": {},
        }
        if self.platform == "google-python":
            config["provider"]["name"] = "google"
            for api in apis:
                config["functions"][api.name] = {
                    "handler": api.name,
                    "events": [{"http": "path"}],
                }
        elif self.platform == "aws-lambda" or self.platform == "aws-lambda-py2":
            config["provider"]["name"] = "aws"
            for api in apis:
                config["functions"][api.name] = {
                    "handler": "handler." + api.name,
                    "events": [{"http": {"path": "/" + api.name, "method": "post"}}],
                }
        else:
            raise BentoMLException(
                "check serverless does not support platform %s at the moment"
                % self.platform
            )
        yaml = YAML()

        with TempDirectory() as tempdir:
            saved_path = os.path.join(tempdir, "serverless.yml")
            yaml.dump(config, Path(saved_path))
            with subprocess.Popen(
                ["serverless", "info"], cwd=tempdir, stdout=PIPE, stderr=PIPE
            ) as proc:
                # We don't use the parse_response function here.
                # Instead of raising error, we will just return false
                content = proc.stdout.read().decode("utf-8")
                response = content.strip().split("\n")
                logger.debug("Serverless response: %s", "\n".join(response))
                error = [s for s in response if "Serverless Error" in s]
                if error:
                    print("has error", "\n".join(response))
                    return False, "\n".join(response)
                else:
                    print("\n".join(response))
                    return True, "\n".join(response)

    def delete(self):
        is_active, _ = self.check_status()
        if not is_active:
            raise BentoMLException(
                "No active deployment for service %s" % self.bento_service.name
            )

        if self.platform == "google-python":
            provider_name = "google"
        elif self.platform == "aws-lambda" or self.platform == "aws-lambda-py2":
            provider_name = "aws"
        config = {
            "service": self.bento_service.name,
            "provider": {
                "name": provider_name,
                "region": self.region,
                "stage": self.stage,
            },
        }
        yaml = YAML()
        with TempDirectory() as tempdir:
            saved_path = os.path.join(tempdir, "serverless.yml")
            yaml.dump(config, Path(saved_path))

            with subprocess.Popen(
                ["serverless", "remove"], cwd=tempdir, stdout=PIPE, stderr=PIPE
            ) as proc:
                response = parse_serverless_response(proc.stdout.read().decode("utf-8"))
                logger.debug("Serverless response: %s", "\n".join(response))
                if self.platform == "google-python":
                    # TODO: Add check for Google's response
                    return True
                elif self.platform == "aws-lambda" or self.platform == "aws-lambda-py2":
                    if "Serverless: Stack removal finished..." in response:
                        return True
                    else:
                        return False
